import random
from copy import copy
import statistics

from typing import List
from jmetal.core.algorithm import Algorithm
from jmetal.util.generator import RandomGenerator
from jmetal.util.termination_criterion import TerminationCriterion
from coco_problem import CocoSolution, CocoProblem
from moarchiving import BiobjectiveNondominatedSortedList as NDA


def strictly_dominates(a, b):
    """Returns True if a << b and False otherwise (assumes 2 objectives)"""
    return (a[0] < b[0]) and (a[1] < b[1])


class BBDEMO(Algorithm[CocoSolution, CocoSolution]):
    """The BB-DEMO algorithm.

    A bi-objective algorithm that creates solutions with BBDE's target-to-best strategy, applies
    perturbation when solutions lose diversity and compares candidates with their parents by their
    uncrowded hypervolume contributions"""
    def __init__(self,
                 problem: CocoProblem,
                 population_size: int,
                 termination_criterion: TerminationCriterion,
                 population_generator=RandomGenerator(),
                 reference_point=None):
        super(BBDEMO, self).__init__()
        self.problem = problem
        self.problem_name = problem.get_name()
        if problem.number_of_objectives != 2:
            raise ValueError('Incorrect number of objectives ({}). Only 2 objective supported.'.
                             format(problem.number_of_objectives))
        if reference_point is None:
            raise ValueError('Reference point for hypervolume is missing')
        self.dimension = int(self.problem.number_of_variables)
        self.number_of_integers = int(self.problem.number_of_integer_variables)
        self.archive = NDA(reference_point=reference_point)
        self.population_generator = population_generator
        self.termination_criterion = termination_criterion
        self.observable.register(termination_criterion)
        self.population_size = population_size
        # Additional algorithm parameter
        self.diversity_threshold = 1e-3
        self.int_perturbation = 3
        self.cont_perturbation = 1e-2

    def create_initial_solutions(self) -> List[CocoSolution]:
        """Fills the initial population of solutions and adds the solutions to the archive"""
        result = []
        for _ in range(self.population_size):
            solution = self.population_generator.new(self.problem)
            result.append(solution)
        return result

    def step(self):
        """Performs one step of the evolution"""
        # Randomize the traversal of the population
        solution_order = list(range(self.population_size))
        random.shuffle(solution_order)
        # Find the best solution
        best = self.get_best_solution()
        # Compute the candidates
        candidates = [self.get_candidate(parent, best) for parent in solution_order]
        # Perturb candidates if diversity is lost
        self.perturb(candidates)
        # Evaluate the candidates
        candidates = [self.problem.evaluate(candidate) for candidate in candidates]
        for parent_i, candidate in zip(solution_order, candidates):
            # Update the parent if the candidate is better
            self.update_if_better(parent_i, candidate)

    def get_candidate(self, parent_i, best) -> CocoSolution:
        """Performs a BBDE-like candidate creation that combines the current solution (the parent)
        with two random solutions and the best solution as follows:
                candidate = parent + F1 * (p_r1 - p_r2) + F2 * (best - parent)
        Takes care of repairing the candidate (not affecting the rotation invariance).
        """
        # Get two random values
        F1 = random.lognormvariate(0.1, 0.5)
        F2 = random.lognormvariate(0.1, 0.5)
        # Get all the individuals needed
        parent = self.solutions[parent_i]
        [r1, r2] = random.sample(self.solutions, 2)
        # Create the candidate
        candidate = copy(parent)
        for i in range(self.dimension):
            # Use the target-to-best strategy
            candidate.variables[i] = parent.variables[i] + \
                                     F1 * (r1.variables[i] - r2.variables[i]) +\
                                     F2 * (best.variables[i] - parent.variables[i])
        # Make sure it stays inside the search space bounds
        candidate.variables = self.repair_solution(parent.variables, candidate.variables,
                                                   parent.lower_bound, parent.upper_bound)
        return candidate

    def repair_solution(self, original, solution, lb, ub):
        """Repairs solution (variables) keeping the direction from the original intact"""
        repair = [1.0] * self.dimension
        needs_repair = False
        for i in range(self.dimension):
            # Repair makes sure that the operator is rotationally invariant
            if (solution[i] < lb[i]) and (original[i] != solution[i]):
                repair[i] = (original[i] - lb[i]) / (original[i] - solution[i])
                needs_repair = True
            elif (solution[i] > ub[i]) and (original[i] != solution[i]):
                repair[i] = (ub[i] - original[i]) / (solution[i] - original[i])
                needs_repair = True
        if needs_repair:
            for i in range(self.dimension):
                solution[i] = original[i] + min(repair) * (solution[i] - original[i])
                # Catch any remaining overflows due to precision errors
                if solution[i] < lb[i]:
                    solution[i] = lb[i]
                elif solution[i] > ub[i]:
                    solution[i] = ub[i]
        for i in range(self.number_of_integers):
            solution[i] = round(solution[i])
        return solution

    def get_best_solution(self):
        """Returns the solution with the best hypervolume contribution """
        hv_contributions = [self.archive.contributing_hypervolume(sol.objectives.tolist())
                            for sol in self.solutions]
        max_hv = max(hv_contributions)
        # Return the solution with the highest hypervolume
        return self.solutions[hv_contributions.index(max_hv)]

    def perturb(self, candidates):
        """Checks candidate diversity for each variable and perturbs candidates in the variables
        that show no diversity"""
        perturb_variables = []
        # Check the diversity for each variable
        for var in range(self.dimension):
            values = [solution.variables[var] for solution in candidates]
            if statistics.stdev(values) <= self.diversity_threshold:
                perturb_variables.append(var)
        if len(perturb_variables) > 0:
            lb = self.solutions[0].lower_bound
            ub = self.solutions[0].upper_bound
            var_range = [ubi - lbi for lbi, ubi in zip(lb, ub)]
            # Handle separately integer and continuous variables
            perturbation_d = [self.int_perturbation if var < self.number_of_integers
                              else var_range[var] * self.cont_perturbation
                              for var in range(self.dimension)]
            # Perform the perturbation (make sure it doesn't go overboard)
            for sol in range(self.population_size):
                center = candidates[sol].variables
                perturbation = [c + random.normalvariate(0, 1) * perturbation_d[var]
                                if var in perturb_variables else c
                                for var, c in enumerate(center)]
                # Repair if needed
                candidates[sol].variables = self.repair_solution(center, perturbation, lb, ub)

    def update_if_better(self, parent_i, candidate):
        """Update the parent if the candidate is better than it"""
        parent_obj = self.solutions[parent_i].objectives.tolist()
        candidate_obj = candidate.objectives.tolist()
        if strictly_dominates(candidate_obj, parent_obj):
            candidate_better = True
        elif strictly_dominates(parent_obj, candidate_obj):
            candidate_better = False
        else:
            parent_hvc = self.archive.contributing_hypervolume(parent_obj)
            temp_archive = self.archive
            if parent_hvc > 0:
                # Remove the parent from the temporary archive
                temp_archive = NDA(self.archive, self.archive.reference_point)
                temp_archive.remove(parent_obj)
            # HV contribution of the candidate w.r.t. the temporary archive
            candidate_hvc = temp_archive.hypervolume_improvement(candidate_obj)
            candidate_better = (candidate_hvc >= parent_hvc)
        if candidate_better:
            self.solutions[parent_i] = candidate
            self.archive.add(candidate.objectives.tolist())

    def stopping_condition_is_met(self) -> bool:
        return self.termination_criterion.is_met

    def get_result(self) -> List[CocoSolution]:
        return self.solutions

    def get_name(self) -> str:
        return 'BBDEMO'

    def evaluate(self, solution_list: List[CocoSolution]) -> List[CocoSolution]:
        for solution in solution_list:
            self.problem.evaluate(solution)
            self.archive.add(solution.objectives)
        return solution_list

    def get_observable_data(self) -> dict:
        return {'EVALUATIONS': self.evaluations}

    def init_progress(self) -> None:
        self.evaluations = self.population_size

        observable_data = self.get_observable_data()
        self.observable.notify_all(**observable_data)

    def update_progress(self) -> None:
        self.evaluations += self.population_size

        observable_data = self.get_observable_data()
        self.observable.notify_all(**observable_data)
