#!/usr/bin/env python
"""
An experiment using the BBDEMO algorithm implemented in JMetal to solve COCO problems.

Arguments:
    suite=SUITE_NAME                   # Suite name (default 'bbob-biobj-mixint')
    suite_options=SUITE_OPTIONS        # Suite options (default '')
    observer=OBSERVER_NAME             # Name of the observer (default 'bbob-biobj')
    observer_options=OBSERVER_OPTIONS  # Observer options (default '')
    budget_multiplier=BUDGET           # Budget multiplier (default 1)
    batches=BATCHES                    # Number of all batches to parallelize the experiment
                                       # (default 1)
    batch=BATCH                        # This batch (default 1)

Example:
    # Runs the 1st of 4 batches of the bbob-biobj-mixint
    `python coco_experiment.py batches=4 batch=1`
"""
import cocoex
import sys
from coco_problem import CocoProblem
from bbdemo import BBDEMO
from jmetal.util.termination_criterion import StoppingByEvaluations
import random


def parse_arguments(argv):
    """Parses the command-line arguments and returns a dictionary with all the arguments"""
    batches = 4
    current_batch = 1
    observer_name = 'bbob-biobj'
    observer_options = ''
    suite_name = 'bbob-biobj-mixint'
    suite_options = 'instance_indices: 1-5'
    budget_multiplier = 2000
    # Parse the command line arguments
    for arg in argv:
        if arg[:6] == 'suite=':
            suite_name = arg[6:]
        elif arg[:14] == 'suite_options=':
            suite_options = arg[14:]
        elif arg[:9] == 'observer=':
            observer_name = arg[9:]
        elif arg[:17] == 'observer_options=':
            observer_options = arg[17:]
        elif arg[:18] == 'budget_multiplier=':
            budget_multiplier = int(arg[18:])
        elif arg[:8] == 'batches=':
            batches = int(arg[8:])
        elif arg[:6] == 'batch=':
            current_batch = int(arg[6:])
    # Print out the options
    print('RW experiment ran with the following options:')
    print('suite = {}'.format(suite_name))
    print('suite_options = {}'.format(suite_options))
    print('observer = {}'.format(observer_name))
    print('observer_options = {}'.format(observer_options))
    print('budget_multiplier = {}'.format(budget_multiplier))
    print('batches = {}'.format(batches))
    print('current_batch = {}'.format(current_batch))
    return dict(suite_name=suite_name, suite_options=suite_options,
                observer_name=observer_name, observer_options=observer_options,
                budget_multiplier=budget_multiplier, batches=batches,
                current_batch=current_batch)


def run_experiment(argv=None):
    # Parse the command-line arguments
    params = parse_arguments(argv)
    suite_name = params['suite_name']
    suite_options = params['suite_options']
    observer_name = params['observer_name']
    observer_options = params['observer_options']
    budget_multiplier = params['budget_multiplier']
    batches = params['batches']
    current_batch = params['current_batch']

    # Prepare the suite
    suite = cocoex.Suite(suite_name, '', suite_options)
    # Prepare the observer
    observer = cocoex.Observer(observer_name, '{} result_folder: {}-{}{} algorithm_name:{}'.format(
        observer_options, suite_name, 'BBDEMO',
        '-batch-{}'.format(current_batch) if batches > 1 else '', 'BBDEMO'))
    # Use minimal printing
    minimal_print = cocoex.utilities.MiniPrint()

    # Run the solver on the problems of the suite in the current batch
    for problem_index, problem in enumerate(suite):
        # Skip problems not in this batch
        if (problem_index + current_batch - 1) % batches:
            continue
        # Observe the problem
        problem.observe_with(observer)
        # Set random seed
        random.seed(69)
        # Use the BBDEMO algorithm
        algorithm = BBDEMO(
            problem=CocoProblem(problem),
            population_size=20,
            termination_criterion=StoppingByEvaluations(
                max_evaluations=budget_multiplier * problem.dimension),
            reference_point=problem.largest_fvalues_of_interest
        )
        algorithm.run()
        # Print progress
        minimal_print(problem, final=problem.index == len(suite) - 1)


if __name__ == '__main__':
    run_experiment(sys.argv)
